from typing import Callable, Union, Any, List, Tuple
from numpy import ndarray, int16, int32, float32, float64
# Note I'm using Tuple to denote fixed sized sequences and List to denote a sliceable unbounded sized iterator and
# tuple as a fixed


from typing import TypeVar, Optional, Iterable, Sequence


def MyType(name: str, constraint, *more_constraints, doc: Optional[str] = None, aka: Optional[Iterable] = None,
           covariant=False, contravariant=False,
           assign_to_globals=False):
    """
    Make a new type with (optional) doc and (optional) aka, set of var names it often appears as

    Args:
        name: Name to give the variable
        constraints: types (see typing.TypeVar)
        doc: Optional string to put in __doc__ attribute
        aka: Optional set (or any iterable) to put in _aka attribute,
            meant to list names the variables of this type often appear as.

    Returns: None

    >>> from typing import Any, List
    >>> T = MyType('T', int)
    >>> type(T)
    <class 'function'>
    >>> Key = MyType('Key', Any, aka=['key', 'k'])
    >>> Key._aka
    {'key', 'k'}
    >>> Val = MyType('Val', int, float, List[Union[int, float]], doc="A number or list of numbers.")
    >>> Val.__doc__
    'A number or list of numbers.'
    """
    if len(more_constraints) == 0:
        new_tp = TypeVar(name, bound=constraint, covariant=covariant, contravariant=contravariant)
    else:
        new_tp = TypeVar(name, constraint, *more_constraints, covariant=covariant, contravariant=contravariant)
    if doc is not None:
        setattr(new_tp, '__doc__', doc)
    if aka is not None:
        setattr(new_tp, '_aka', set(aka))
    if assign_to_globals:
        globals()[name] = new_tp  # not sure how kosher this is... Should only use at top level of module, for sure!
    return new_tp


# This would have been convenient, but pycharm doesn't see the globals created by NT!
# from functools import partial
# NT = partial(new_type, assign_to_globals=True)


FixedSizeSeq = MyType('FixedSizeSeq', Tuple, List, ndarray)

VarSizeSeq = MyType('VarSizeSeq', List, ndarray)

Key = MyType('Key', Any, doc="Any object used to reference another", aka={'key', 'k'})

# Waveform = Iterable[Union[float, int]]
Sample = MyType('Sample', float, int, int16, int32, float32, float64,
                doc="The numerical value of a digital signal sample")

# Waveform = MyType('Waveform', VarSizeSeq[Sample])
Waveform = MyType('Waveform', Sequence[Sample])
Waveforms = Iterable[Waveform]

# WfGen = MyType('WfGen', Iterable[Waveform], doc='A iterable of Waveforms')
KeyWfGen = MyType('KeyWfGen', Iterable[Tuple[Key, Waveform]],
                  doc='A iterable of (Key, Waveform) pairs')

# Chunk = MyType('Chunk', FixedSizeSeq[Sample])
Chunk = MyType('Chunk', Sequence[Sample], aka=['chk'])
Chunks = MyType('Chunks', Iterable[Chunk], aka=['chunks', 'chks'])
Chunker = MyType('Chunker', Callable[[Waveform], Iterable[Chunk]], aka=['chunker', 'wf_to_chks'],
                 doc='The component that generates Chunks from a Waveform')

Feature = MyType('Feature', float, int, int16, int32, float32, float64,
                 doc="A number that represents a caracteristic of something. "
                     "Usually appears as an item of an FV (a sequence of Features)")
# FV = FixedSizeSeq[Feature]

FV = MyType('FV', Sequence[Feature], doc="Feature Vector. The informational fingerprint of something.")
FVs = MyType('FVs', Iterable[FV], aka=['fvs'])
Featurizer = MyType('Featurizer', Callable[[Any], FV],
                    doc="A function that makes FVs (out of Chunks, other FVs, or anything really. "
                        "(This is a declaration that the output will be FVs, not what the input should be.)")
ChkFeaturizer = MyType('ChkFeaturizer', Callable[[Chunk], FV], aka=['featurizer', 'chk_to_fv'],
                       doc="A function that makes FVs specifically from Chunks.")
# ChkFeaturizer
# Note: Snips are ints, but with an upper limit... From an "alphabet", so akin to a very big Enum in a sense.
Snip = MyType('Snip', int, aka=['snip'],
              doc="The smallest element of a signal language. "
                  "Technically, an index representing a region of a feature space partition.")
Snips = MyType('Snips', Iterable[Snip], aka=['snips'],
               doc="A sequence or stream whose elements are Snips")

Quantizer = MyType('Quantizer', Callable[[Any], Snip], aka=['quantizer', 'fv_to_snip'],
                   doc='The function that computes a Snip out of an FV.')

Snipper = MyType('Snipper', Callable[[Waveform], Snips], aka=['snipper'],
                 doc="The function that gets you from a stream of samples (Waveform) to "
                     "a stream of snips (Snips)")



from slang.core import KvDataSource, Snipper
from slang.snippers import ClassificationSnipper
from slang.util import snips_to_str as dflt_snips_to_str
from slang.util import ddir
from slang.chunkers import fixed_step_chunker



from typing import Callable
from slang.chunkers import fixed_step_chunker
from sklearn.decomposition import PCA
from sklearn.cluster import KMeans
from slang.snip_stats import BayesFactors
from slang.core import Snipper


class DfltWfToChk:
    def __init__(self, chk_size=2048, chk_step=None):
        if chk_step is None:
            chk_step = chk_size
        self.chk_size = chk_size
        self.chk_step = chk_step

    def fit(self, *args, **kwargs):
        return self  # no fitting required

    def __call__(self, wf):
        yield from fixed_step_chunker(wf, chk_size=self.chk_size, chk_step=self.chk_step)


class DfltChkToFv(PCA):
    def __init__(self, n_components=5, *args, **kwargs):
        super().__init__(n_components=n_components, *args, **kwargs)

    def __call__(self, fv):
        return self.transform([fv])[0]


class DfltFvToSnip(KMeans):
    def __init__(self, n_clusters=47, *args, **kwargs):
        super().__init__(n_clusters=n_clusters, *args, **kwargs)

    def __call__(self, fv):
        return self.predict([fv])[0]


def is_iterable(o):
    try:
        iter(o)
        return True
    except TypeError:
        return False


def _get_pairs(iterables):
    """ """
    if hasattr(iterables, '__len__'):
        if len(iterables) == 2:
            return iterables
        elif len(iterables) == 1 and is_iterable(iterables[0]):
            return list(zip(*iterables[0]))
    else:
        return list(zip(*iterables))


def _is_a_tuple_of_aligned_iterables(x):
    return hasattr(x, '__len__') and len(x) == 2


def iterate_over_pairs(pairs):
    if _is_a_tuple_of_aligned_iterables(pairs):
        yield from zip(*pairs)
    else:
        yield from pairs  # assume pairs is an iterable of pairs


class ClassificationSnipper(Snipper):
    wf_to_chks: Callable
    chk_to_fv: Callable  # TODO: Make a "Callable and Fittable" type
    fv_to_snip: Callable  # TODO: Make a "Callable and Fittable" type
    snip_to_score: Callable  # TODO: Make a "Callable and Fittable" type

    dflt_cls_of_name = {
        'wf_to_chks': DfltWfToChk,
        'chk_to_fv': DfltChkToFv,
        'fv_to_snip': DfltFvToSnip,
        'snip_to_score': BayesFactors
    }

    @classmethod
    def mk_component(cls, obj, role=None):
        if isinstance(obj, type):
            return obj()
        elif not callable(obj):
            assert isinstance(obj, dict)
            return cls.dflt_cls_of_name.get(role)(**obj)
        else:
            return obj  # as is

    def __init__(self, wf_to_chks=DfltWfToChk(),
                 chk_to_fv=DfltChkToFv,
                 fv_to_snip=DfltFvToSnip,
                 snip_to_score=BayesFactors):

        for obj_role in ['wf_to_chks', 'chk_to_fv', 'fv_to_snip', 'snip_to_score']:
            obj = self.mk_component(locals()[obj_role], obj_role)
            assert callable(obj)
            setattr(self, obj_role, obj)

        # Note: Equivalent to below:
        # self.wf_to_chks = self.mk_component(wf_to_chks, 'wf_to_chks')
        # self.chk_to_fv = self.mk_component(chk_to_fv, 'chk_to_fv')
        # self.fv_to_snip = self.mk_component(fv_to_snip, 'fv_to_snip')
        # self.snip_to_score = self.mk_component(snip_to_score, 'snip_to_score')

        assert hasattr(self.chk_to_fv, 'fit')
        assert hasattr(self.fv_to_snip, 'fit')
        assert hasattr(self.snip_to_score, 'fit')

    def fit_chk_to_fv(self, *chks_tags):
        chks, tags = _get_pairs(chks_tags)
        self.chk_to_fv.fit(chks, tags)

    def fit_fv_to_snip(self, *fvs_tags):
        fvs, tags = _get_pairs(fvs_tags)
        self.fv_to_snip.fit(fvs, tags)

    def fit_snip_to_score(self, *snips_tags):
        snips, tags = _get_pairs(snips_tags)
        self.snip_to_score.fit(snips, tags)

    # def snips_of_wf(self, wf: Waveform) -> Snips:
    #     warn("The name 'snips_of_wf' be replaced by 'wf_to_snips' soon.")
    #     for chk in self.wf_to_chks(wf):
    #         fv = self.chk_to_fv(chk)
    #         yield self.fv_to_snip(fv)
    #
    # def wf_to_fvs(self, wf: Waveform) -> FVs:
    #     for chk in self.wf_to_chks(wf):
    #         yield self.chk_to_fv(chk)
    #
    # def chk_to_snip(self, chk: Chunk) -> Snip:
    #     return self.fv_to_snip(self.chk_to_fv(chk))
    #
    # def wf_to_snips(self, wf: Waveform) -> Snips:
    #     for chk in self.wf_to_chks(wf):
    #         fv = self.chk_to_fv(chk)
    #         yield self.fv_to_snip(fv)


SlangClassifier = ClassificationSnipper  # alias for back compatibility



from setuptools import setup, find_packages


def readme():
    with open('README.md') as f:
        return f.read()


setup(
    name='slang',
    version='0.0.4',
    description='Tools to endow signals with language-like structure.',
    long_description=readme(),
    long_description_content_type="text/markdown",
    url='https://github.com/thorwhalen/slang',
    author='Thor Whalen',
    license='Apache',
    packages=find_packages(),
    install_requires=['numpy', 'pandas'],
    include_package_data=True,
    zip_safe=False,
    # download_url='https://github.com/i2mint/py2store/archive/v0.0.5.zip',
    keywords=['sound recognition', 'machine learning', 'language', 'audio', 'signal processing',
              'natural language processing', 'NLP', 'text mining'],
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Topic :: Software Development',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python :: 3.7',
    ],
)



from matplotlib.pylab import *
from numpy import *
from collections import Counter
import pandas as pd
from itertools import chain
from slang.core import DFLT_CHUNKER
# from agen.sound.audio import Sound
from slang.featurizers import _rms_and_crossing_levels, _zero_crossing_ratio
import seaborn as sns


def zero_crossing_count(wf):
    return sum(diff(array(wf) > 0).astype(int))


def heatmap_of_rms_and_crossing_levels(wf_gen, chunker=DFLT_CHUNKER):
    x = list()
    y = list()
    for wf in wf_gen:
        for chk in chunker(wf):
            xx, yy = _rms_and_crossing_levels(chk)
            x.append(xx)
            y.append(yy)

    t = pd.DataFrame({'rms': x, 'zero_crossings': y}).groupby(['rms', 'zero_crossings']).size()
    # t.plot(x='vol', y='zero_crossings', kind='scatter', figsize=(9, 9), alpha=0.2)
    t = t.unstack()
    figure(figsize=(11, 9));
    sns.heatmap(t.T, cmap='gray_r');


def hist_of_zero_crossings(wf_gen, chunker=DFLT_CHUNKER):
    c = Counter(zero_crossing_count(chk) for chk in chain(*map(chunker, wf_gen)))

    pd.Series(c).sort_index().plot(kind='bar', figsize=(16, 5))
    plt.xticks(range(max(c.keys())), range(max(c.keys())))
    plt.xlabel('zero crossings');
    plt.ylabel('count')


def rms_zcr_scatter(wf_gen, chunker=DFLT_CHUNKER):
    def gen(wf_gen):
        for wf in wf_gen:
            for chk in chunker(wf):
                yield std(chk), _zero_crossing_ratio(chk)

    t = pd.DataFrame([{'rms': rms, 'zero_crossing_ratio': zcr} for rms, zcr in gen(wf_gen)])
    t.plot(x='rms', y='zero_crossing_ratio', kind='scatter', figsize=(9, 9), alpha=0.2)
